@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2004-2024 ZNC, see the NOTICE file for details.
+ * Copyright (C) 2004-2025 ZNC, see the NOTICE file for details.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -216,6 +216,10 @@ class CFailToBanMod : public CModule {
             Add(sRemoteIP, 1);
     }
 
+    void OnClientLogin() override {
+        Remove(GetClient()->GetRemoteIP());
+    }
+
     EModRet OnLoginAttempt(std::shared_ptr<CAuthBase> Auth) override {
         // e.g. webadmin ends up here
         const CString& sRemoteIP = Auth->GetRemoteIP();
