@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2004-2024 ZNC, see the NOTICE file for details.
+ * Copyright (C) 2004-2025 ZNC, see the NOTICE file for details.
  * Copyright (C) 2006-2007, CNU <bshalm@broadpark.no>
  *(http://cnu.dieplz.net/znc)
  *
@@ -439,10 +439,14 @@ void CLogMod::OnKick(const CNick& OpNick, const CString& sKickedNick,
 void CLogMod::OnQuit(const CNick& Nick, const CString& sMessage,
                      const vector<CChan*>& vChans) {
     if (NeedQuits()) {
-        for (CChan* pChan : vChans)
+        for (CChan* pChan : vChans) {
+            // Core calls this only for enabled channels, but
+            // OnSendToIRCMessage() below calls OnQuit() for all channels.
+            if (pChan->IsDisabled()) continue;
             PutLog("*** Quits: " + Nick.GetNick() + " (" + Nick.GetIdent() +
                        "@" + Nick.GetHost() + ") (" + sMessage + ")",
                    *pChan);
+        }
     }
 }
 
