@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2004-2024 ZNC, see the NOTICE file for details.
+ * Copyright (C) 2004-2025 ZNC, see the NOTICE file for details.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -160,6 +160,17 @@ class ZNC_EXPORT_LIB_EXPORT CPerlModule : public CModule {
     EModRet OnTopicMessage(CTopicMessage& Message) override;
     EModRet OnSendToClientMessage(CMessage& Message) override;
     EModRet OnSendToIRCMessage(CMessage& Message) override;
+    EModRet OnUserTagMessage(CTargetMessage& Message) override;
+    EModRet OnChanTagMessage(CTargetMessage& Message) override;
+    EModRet OnPrivTagMessage(CTargetMessage& Message) override;
+    EModRet OnInviteMessage(CInviteMessage& Message) override;
+
+    void OnClientGetSASLMechanisms(SCString& ssMechanisms) override;
+    EModRet OnClientSASLServerInitialChallenge(const CString& sMechanism,
+                                               CString& sResponse) override;
+    EModRet OnClientSASLAuthenticate(const CString& sMechanism,
+                                     const CString& sMessage) override;
+    void OnClientSASLAborted() override;
 };
 
 static inline CPerlModule* AsPerlModule(CModule* p) {
