@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2004-2024 ZNC, see the NOTICE file for details.
+ * Copyright (C) 2004-2025 ZNC, see the NOTICE file for details.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -118,7 +118,7 @@ bool CNick::RemPerm(char cPerm) {
 
 char CNick::GetPermChar() const {
     CIRCSock* pIRCSock = (!m_pNetwork) ? nullptr : m_pNetwork->GetIRCSock();
-    const CString& sChanPerms = (!pIRCSock) ? "@+" : pIRCSock->GetPerms();
+    const CString sChanPerms = (!pIRCSock) ? "@+" : pIRCSock->GetPerms();
 
     for (unsigned int a = 0; a < sChanPerms.size(); a++) {
         const char& c = sChanPerms[a];
@@ -132,7 +132,7 @@ char CNick::GetPermChar() const {
 
 CString CNick::GetPermStr() const {
     CIRCSock* pIRCSock = (!m_pNetwork) ? nullptr : m_pNetwork->GetIRCSock();
-    const CString& sChanPerms = (!pIRCSock) ? "@+" : pIRCSock->GetPerms();
+    const CString sChanPerms = (!pIRCSock) ? "@+" : pIRCSock->GetPerms();
     CString sRet;
 
     for (unsigned int a = 0; a < sChanPerms.size(); a++) {
