@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2004-2024 ZNC, see the NOTICE file for details.
+ * Copyright (C) 2004-2025 ZNC, see the NOTICE file for details.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -293,6 +293,16 @@ TEST_F(ClientTest, OnUserNoticeMessage) {
 }
 
 TEST_F(ClientTest, OnUserJoinMessage) {
+    // JOIN only is sent to server after registration is complete, so make sure
+    // it is complete. But it requires clearing the state afterwards.
+    m_pTestSock->ReadLine("001");
+    m_pTestModule->vsHooks.clear();
+    m_pTestModule->vsMessages.clear();
+    m_pTestModule->vNetworks.clear();
+    m_pTestModule->vClients.clear();
+    m_pTestModule->vChannels.clear();
+    m_pTestSock->vsLines.clear();
+
     CMessage msg("JOIN #chan key");
     m_pTestModule->eAction = CModule::HALT;
     m_pTestClient->ReadLine(msg.ToString());
